/*
 * Copyright 2013-2018 Radian Memory Systems, Inc.  All Rights Reserved.
 *
 * This file is provided under non-disclosure agreement.
 * See the file COPYING for details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <asm-generic/types.h>
#include <asm-generic/ioctl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <errno.h>
#include <asm-generic/types.h>
typedef __u8 u8;
typedef __u16 u16;
typedef __u32 u32;
typedef __u64 u64;
#include "nvme_express.h"
#include "nvme_ioctl.h"
#define MAX_DIES 256
#include "RMSVendorSpecific.h"
#include "nvme_util.h"
#include "dialogDef.h"

#define CONTROL_DEV		"/dev/nvme0"
#define MAX_BOARD_NAME		256

static int
issueVendorSpecific(int fd, __u32 *data)
{
	struct nvme_cmd	*pCmd;
	struct usr_io *pUio;
	int	status;

	pUio = (struct usr_io *) malloc(sizeof(struct usr_io));
	if (pUio == NULL) {
		printf("%s:%d: Could not allocate memory\n",
			__func__, __LINE__);
		return -1;
	}

	/*
	* Initialize the nvme_cmd structure to zero.
	*/
	memset(pUio, 0x0, sizeof(struct usr_io));

	pUio->direction = XFER_TO_DEV;
	pUio->length = 0;
	pUio->addr = (__u64) 0;
	pUio->timeout = 3;
	pCmd = &pUio->cmd;
	/*
	* To fill the USER_NVM_CMD structure parameters with proper values.
	*/
	pCmd->header.opCode = NVME_VNDR_CMD_SYS_DEBUG;
	pCmd->header.cmdID = 0;
	pCmd->header.fusedOp = 0;
	pCmd->header.namespaceID = 0;
	pCmd->header.metadataPtr = 0;
	pCmd->header.prp[0].addr = 0;
	pCmd->header.prp[1].addr = 0;
	pCmd->cmd.vendorSpecific.buffNumDW = 0;
	pCmd->cmd.vendorSpecific.metaNumDW = 0;
	pCmd->cmd.vendorSpecific.vndrCDW12 = (__u32)SYS_DEBUG_DIALOG_CTRL;
	pCmd->cmd.vendorSpecific.vndrCDW13 = data[0];
	pCmd->cmd.vendorSpecific.vndrCDW14 = data[1];
	pCmd->cmd.vendorSpecific.vndrCDW15 = data[2];

	status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, (__u8 *)pCmd);

	if (status != 0) {
		printf("%s:%d: Status %d\n",
			__func__, __LINE__,
			status);
	}

	if (pUio->comp.SCT != 0 || pUio->comp.SC != 0) {
		printf("Operation failed, NVMe error = %s\n",
			nvme_err(pUio->comp.SCT, pUio->comp.SC));
		goto done;
	}

done:

	free(pUio);

	return status;
}

static void
usage(void)
{
	printf("   Controls the behaviour of the dialog software\n");
	printf("usage:  dialog_ctrl [devName] mode\n");
	printf("   devName: The device to control - default /dev/nvme0\n");
	printf("   mode must be one of\n");
	printf("   -c          Clears the dialog page (requires reboot)\n");
	printf("   -e <bit>    Inject an error condition\n");
	printf("   -m <sample> Do capacitance/resistance measurment\n");
	printf("   -p <code>   Set Production code\n");
	printf("   -s          Take a snapshot\n");
	printf("   -t          Sets the current time\n");
}

static void
wait_for_q(void)
{
	system("/bin/stty raw");
	getchar();
	system("/bin/stty cooked");
}

int
main(int argc, char *argv[])
{
	int	c, fd;
	int	flag = 0;
	__u32	ctrlData[3];
	char	*control_dev = CONTROL_DEV;
	char	*pEnv;

	/* Override default device */
	if (argc >= 2 && strncmp("/dev/", argv[1], 5) == 0) {
		control_dev = argv[1];
	} else {
		pEnv = getenv("RMS_DEVICE");
		if (pEnv) {
            control_dev = pEnv;
		}
	}

	memset(ctrlData, 0, sizeof(ctrlData));

	while (1) {

		c = getopt(argc, argv, "ce:m:p:st");

		if (c == -1)
			break;

		switch (c) {

		case 'c':
			ctrlData[0] = CLEAR_DIALOG_PAGE;
			flag++;
			break;

		case 'e':
			ctrlData[0] = SET_ERROR_CONDITION;
			ctrlData[1] = atoi(optarg);
			flag++;
			break;

		case 'm':
			ctrlData[0] = DO_CAP_ESR_MEASUREMENT_CODE;
			ctrlData[1] = atoi(optarg);
			flag++;
			break;

		case 'p':
			ctrlData[0] = ENTER_PRODUCTION_CODE;
			ctrlData[1] = atoi(optarg);
			flag++;
			break;

		case 's':
			ctrlData[0] = TAKE_A_SNAP_SHOT;
			flag++;
			break;

		case 't':
			ctrlData[0] = SET_TIME_IN_SECONDS;
			ctrlData[1] = (u32)time(NULL);
			flag++;
			break;

		case 'H':
			usage();
			wait_for_q();
			exit(1);
			break;

		case 'h':
		default:
			usage();
			exit(1);
		}
	}

	if (flag > 1) {
		printf("Only one of -c -m -p -s or -t\n");
		usage();
		exit(1);
	}

	if (flag == 0) {
		usage();
		exit(1);
	}

	fd = open(control_dev, O_RDWR);
	if (fd < 0) {
		printf("Could not open file %s ", control_dev);
		if (errno == EACCES)
			printf(" Access error - try running as su\n");
		else
			printf(" Error %d\n", errno);
		return -1;
	}

	if (issueVendorSpecific(fd, ctrlData) < 0)
		printf("Failed issuing vendor specific command\n");

	close(fd);
	return 0;
}

